@@ -8,7 +8,17 @@ use warnings;
 use Carp;
 use Encode ();
 use List::Util ();
-use PPIx::QuoteLike::Constant qw{ MINIMUM_PERL VARIABLE_RE @CARP_NOT };
+use PPIx::QuoteLike::Constant qw{
+    ARRAY_REF
+    LOCATION_LINE
+    LOCATION_CHARACTER
+    LOCATION_COLUMN
+    LOCATION_LOGICAL_LINE
+    LOCATION_LOGICAL_FILE
+    MINIMUM_PERL
+    VARIABLE_RE
+    @CARP_NOT
+};
 use PPIx::QuoteLike::Token::Control;
 use PPIx::QuoteLike::Token::Delimiter;
 use PPIx::QuoteLike::Token::Interpolation;
@@ -16,9 +26,29 @@ use PPIx::QuoteLike::Token::String;
 use PPIx::QuoteLike::Token::Structure;
 use PPIx::QuoteLike::Token::Unknown;
 use PPIx::QuoteLike::Token::Whitespace;
+use PPIx::QuoteLike::Utils qw{
+    column_number
+    line_number
+    logical_filename
+    logical_line_number
+    statement
+    visual_column_number
+    __instance
+    __match_enclosed
+    __matching_delimiter
+};
 use Scalar::Util ();
+use Text::Tabs ();
 
-our $VERSION = '0.008';
+our $VERSION = '0.018';
+
+use constant CLASS_CONTROL       => 'PPIx::QuoteLike::Token::Control';
+use constant CLASS_DELIMITER     => 'PPIx::QuoteLike::Token::Delimiter';
+use constant CLASS_INTERPOLATION => 'PPIx::QuoteLike::Token::Interpolation';
+use constant CLASS_STRING        => 'PPIx::QuoteLike::Token::String';
+use constant CLASS_STRUCTURE     => 'PPIx::QuoteLike::Token::Structure';
+use constant CLASS_UNKNOWN       => 'PPIx::QuoteLike::Token::Unknown';
+use constant CLASS_WHITESPACE    => 'PPIx::QuoteLike::Token::Whitespace';
 
 use constant CODE_REF	=> ref sub {};
 
@@ -26,26 +56,45 @@ use constant ILLEGAL_FIRST	=>
     'Tokenizer found illegal first characters';
 use constant MISMATCHED_DELIM	=>
     'Tokenizer found mismatched delimiters';
-
-$PPIx::QuoteLike::DEFAULT_POSTDEREF = 1;
+use constant NO_INDENTATION	=>
+    'No indentation string found';
 
 {
-    my $match_sq = _match_enclosed( qw< ' > );
-    my $match_dq = _match_enclosed( qw< " > );
-    my $match_bt = _match_enclosed( qw< ` > );
+    my $match_sq = __match_enclosed( qw< ' > );
+    my $match_dq = __match_enclosed( qw< " > );
+    my $match_bt = __match_enclosed( qw< ` > );
 
     sub new {	## no critic (RequireArgUnpacking)
 	my ( $class, $source, %arg ) = @_;
 
 	my @children;
 
-	defined $arg{postderef}
-	    or $arg{postderef} = $PPIx::QuoteLike::DEFAULT_POSTDEREF;
+	if ( defined $arg{postderef} ) {
+	    $class->_deprecation_notice( attribute => 'postderef' );
+	} else {
+	    $arg{postderef} = 1;
+	}
+
+	if ( $arg{location} ) {
+	    ARRAY_REF eq ref $arg{location}
+		or croak q<Argument 'location' must be an array reference>;
+	    foreach my $inx ( 0 .. 3 ) {
+		$arg{location}[$inx] =~ m/ [^0-9] /smx
+		    and croak "Argument 'location' element $inx must be an unsigned integer";
+	    }
+	}
+
+	if ( ! defined $arg{index_locations} ) {
+	    $arg{index_locations} = !! $arg{location} ||
+		__instance( $source, 'PPI::Element' );
+	}
 
 	my $self = {
+	    index_locations	=> $arg{index_locations},
 	    children	=> \@children,
 	    encoding	=> $arg{encoding},
 	    failures	=> 0,
+	    location	=> $arg{location},
 	    postderef	=> ( $arg{postderef} ? 1 : 0 ),
 	    source	=> $source,
 	};
@@ -55,7 +104,7 @@ $PPIx::QuoteLike::DEFAULT_POSTDEREF = 1;
 	defined( my $string = $self->_stringify_source( $source ) )
 	    or return;
 
-	my ( $type, $gap, $content, $end_delim, $start_delim );
+	my ( $type, $gap, $gap2, $content, $end_delim, $indented, $start_delim );
 
 	$arg{trace}
 	    and warn "Initial match of $string\n";
@@ -65,14 +114,14 @@ $PPIx::QuoteLike::DEFAULT_POSTDEREF = 1;
 	    ( $type, $gap, $start_delim ) = ( $1, $2, $3 );
 	    not $gap
 		and $start_delim =~ m< \A \w \z >smx
-		and return $self->_link_elems( $self->_unknown(
-		    $string, ILLEGAL_FIRST ) );
+		and return $self->_link_elems( $self->_make_token(
+		    CLASS_UNKNOWN, $string, error => ILLEGAL_FIRST ) );
 	    $arg{trace}
 		and warn "Initial match '$type$start_delim'\n";
 	    $self->{interpolates} = 'qq' eq $type ||
 		'qx' eq $type && q<'> ne $start_delim;
 	    $content = substr $string, ( pos $string || 0 );
-	    $end_delim = _matching_delimiter( $start_delim );
+	    $end_delim = __matching_delimiter( $start_delim );
 	    if ( $end_delim eq substr $content, -1 ) {
 		chop $content;
 	    } else {
@@ -83,33 +132,40 @@ $PPIx::QuoteLike::DEFAULT_POSTDEREF = 1;
 	# Note that the regexp used here is slightly wrong in that white
 	# space between the '<<' and the termination string is not
 	# allowed if the termination string is not quoted in some way.
-	} elsif ( $string =~ m/ \A \s* ( << ) ( \s* )
-	    ( \w+ | $match_sq | $match_dq | $match_bt ) \n /smxgc ) {
-	    ( $type, $gap, $start_delim ) = ( $1, $2, $3 );
+	} elsif ( $string =~ m/ \A \s* ( << ) ( \s* ) ( ~? ) ( \s* )
+	    ( [\\]? \w+ | $match_sq | $match_dq | $match_bt ) \n /smxgc ) {
+	    ( $type, $gap, $indented, $gap2, $start_delim ) = (
+		$1, $2, $3, $4, $5 );
 	    $arg{trace}
-		and warn "Initial match '$type$start_delim$gap'\n";
-	    $self->{interpolates} = $start_delim !~ m/ \A ' /smx;
+		and warn "Initial match '$type$start_delim$gap$indented'\n";
+	    $self->{interpolates} = $start_delim !~ m/ \A [\\'] /smx;
 	    $content = substr $string, ( pos $string || 0 );
 	    $end_delim = _unquote( $start_delim );
-	    if ( $content =~ s/ ^ \Q$end_delim\E \n? \z //smx ) {
+	    # NOTE that the indentation is specifically space or tab
+	    # only.
+	    if ( $content =~ s/ ^ ( [ \t]* ) \Q$end_delim\E \n? \z //smx ) {
+		# NOTE PPI::Token::HereDoc does not preserve the
+		# indentation of an indented here document, so the
+		# indentation will appear to be '' if we came from PPI.
+		if ( $indented ) {
+		    # Version per perldelta.pod for that release.
+		    $self->{perl_version_introduced} = '5.025007';
+		    $self->{indentation} = "$1";
+		    $self->{_indentation_re} = qr/
+			^ \Q$self->{indentation}\E /smx;
+		}
 	    } else {
 		$end_delim = '';
 	    }
 	    $self->{start} = [
-		PPIx::QuoteLike::Token::Delimiter->__new(
-		    content	=> $start_delim,
-		),
-		PPIx::QuoteLike::Token::Whitespace->__new(
-		    content	=> "\n",
-		),
+		$self->_make_token( CLASS_DELIMITER, $start_delim ),
+		$self->_make_token( CLASS_WHITESPACE, "\n" ),
 	    ];
+
+	    # Don't instantiate yet -- we'll do them at the end.
 	    $self->{finish} = [
-		PPIx::QuoteLike::Token::Delimiter->__new(
-		    content	=> $end_delim,
-		),
-		PPIx::QuoteLike::Token::Whitespace->__new(
-		    content	=> "\n",
-		),
+		[ CLASS_DELIMITER, $end_delim ],
+		[ CLASS_WHITESPACE, "\n" ],
 	    ];
 
 	# ``, '', "", <>
@@ -119,7 +175,7 @@ $PPIx::QuoteLike::DEFAULT_POSTDEREF = 1;
 		and warn "Initial match '$type$start_delim'\n";
 	    $self->{interpolates} = q<'> ne $start_delim;
 	    $content = substr $string, ( pos $string || 0 );
-	    $end_delim = _matching_delimiter( $start_delim );
+	    $end_delim = __matching_delimiter( $start_delim );
 	    if ( $end_delim eq substr $content, -1 ) {
 		chop $content;
 	    } else {
@@ -130,109 +186,105 @@ $PPIx::QuoteLike::DEFAULT_POSTDEREF = 1;
 	} else {
 	    $arg{trace}
 		and warn "No initial match\n";
-	    return $self->_link_elems( $self->_unknown(
-		    $string, ILLEGAL_FIRST ) );
+	    return $self->_link_elems( $self->_make_token(
+		    CLASS_UNKNOWN, $string, error => ILLEGAL_FIRST ) );
 	}
 
 	$self->{interpolates} = $self->{interpolates} ? 1 : 0;
 
 	$self->{type} = [
-	    PPIx::QuoteLike::Token::Structure->__new(
-		content	=> $type,
-	    ),
-	    length $gap ? PPIx::QuoteLike::Token::Whitespace->__new(
-		content	=> $gap,
-	    ) : ()
+	    $self->_make_token( CLASS_STRUCTURE, $type ),
+	    length $gap ?
+		$self->_make_token( CLASS_WHITESPACE, $gap ) :
+		(),
+	    length $indented ?
+	        $self->_make_token( CLASS_STRUCTURE, $indented ) :
+		(),
+	    length $gap2 ?
+		$self->_make_token( CLASS_WHITESPACE, $gap2 ) :
+		(),
 	];
 	$self->{start} ||= [
-	    PPIx::QuoteLike::Token::Delimiter->__new(
-		content	=> $start_delim,
-	    ),
+	    $self->_make_token( CLASS_DELIMITER, $start_delim ),
 	];
 
 	$arg{trace}
 	    and warn "Without delimiters: '$content'\n";
 
+	# We accumulate data and manufacure tokens at the end to reduce
+	# the overhead involved in merging strings.
 	if ( $self->{interpolates} ) {
-	    {	# Single-iteration loop
+	    push @children, [ '' => '' ];	# Prime the pump
+	    while ( 1 ) {
 
 		if ( $content =~ m/ \G ( \\ [ULulQEF] ) /smxgc ) {
-		    push @children, PPIx::QuoteLike::Token::Control->__new(
-			content	=> "$1",		# Remove magic
-		    );
-		    redo;
-		}
-
-		# Handle \N{...} separately because it can not contain
-		# an interpolation even inside of an
-		# otherwise-interpolating string. That is to say,
-		# "\N{$foo}" is simply invalid, and does not even try to
-		# interpolate $foo.  {
-		if ( $content =~ m/ \G ( \\ N [{] ( [^}]+ ) [}] ) /smxgc ) {
+		    push @children, [ CLASS_CONTROL, "$1" ];
+		} elsif (
+		    $content =~ m/ \G ( \\ N [{] ( [^}]+ ) [}] ) /smxgc
+		) {
+		    # Handle \N{...} separately because it can not
+		    # contain an interpolation even inside of an
+		    # otherwise-interpolating string. That is to say,
+		    # "\N{$foo}" is simply invalid, and does not even
+		    # try to interpolate $foo.  {
+		    # TODO use $re = __match_enclosed( '{' ); # }
 		    my ( $seq, $name ) = ( $1, $2 );
 		    # TODO The Regexp is certainly too permissive. For
 		    # the moment all I am doing is disallowing
 		    # interpolation.
 		    push @children, $name =~ m/ [\$\@] /smx ?
-			$self->_unknown( $seq, "Unknown charname '$name'" ) :
-			PPIx::QuoteLike::Token::String->__new(
-			    content	=> $seq,
-			);
-		    redo;
-		}
-
-		if ( $content =~ m/ \G ( [\$\@] \#? \$* ) /smxgc ) {
+			[ CLASS_UNKNOWN, $seq,
+			    error => "Unknown charname '$name'" ] :
+			[ CLASS_STRING, $seq ];
+		# NOTE in the following that I do not read perldata as
+		# saying there can be space between the sigil and the
+		# variable name, but Perl itself seems to accept it as
+		# of 5.30.1.
+		} elsif ( $content =~ m/ \G ( [\$\@] \#? \$* ) /smxgc ) {
 		    push @children, $self->_interpolation( "$1", $content );
-		    redo;
-		}
-
-		if ( $content =~ m/ \G ( \\ . | [^\\\$\@]+ ) /smxgc ) {
-		    my $content = $1;
-		    @children
-			and $children[-1]->isa(
-			    'PPIx::QuoteLike::Token::String' )
-			and $content = ( pop @children )->content() .
-		    $content;
-		    push @children, PPIx::QuoteLike::Token::String->__new(
-			content	=> $content,
-		    );
-		    redo;
-		}
-	    }
-
-	    # We might have consecutive strings if _interpolation()
-	    # generated a string rather than an interpolation. Merge
-	    # these.
-	    my @rslt;
-	    foreach my $elem ( @children ) {
-		if ( $elem->isa( 'PPIx::QuoteLike::Token::String' ) &&
-		    @rslt &&
-		    $rslt[-1]->isa( 'PPIx::QuoteLike::Token::String' )
-		) {
-		    push @rslt, PPIx::QuoteLike::Token::String->__new(
-			content	=> join( '', map { $_->content() } pop
-			    @rslt, $elem ),
+		} elsif ( $content =~ m/ \G ( \\ . | [^\\\$\@]+ ) /smxgc ) {
+		    push @children, $self->_remove_here_doc_indentation(
+			"$1",
+			sibling	=> \@children,
 		    );
 		} else {
-		    push @rslt, $elem;
+		    last;
 		}
 	    }
-	    @children = @rslt;
+
+	    @children = _merge_strings( @children );
+	    shift @children;	# remove the priming
+
+	    # Make the tokens, at long last.
+	    foreach ( @children ) {
+		$_ = $self->_make_token( @{ $_ } );
+	    }
 
 	} else {
 
 	    length $content
-		and push @children, PPIx::QuoteLike::Token::String->__new(
-		    content	=> $content,
-		);
+		and push @children, map { $self->_make_token( @{ $_ } ) }
+		    _merge_strings(
+			$self->_remove_here_doc_indentation( $content )
+			);
 
 	}
 
-	$self->{finish} ||= [
-	    PPIx::QuoteLike::Token::Delimiter->__new(
-		content	=> $end_delim,
-	    ),
-	];
+	# Add the indentation before the end marker, if needed
+	$self->{indentation}
+	    and push @children, $self->_make_token(
+	    CLASS_WHITESPACE, $self->{indentation} );
+
+	if ( $self->{finish} ) {
+	    # If we already have something here it is data, not objects.
+	    foreach ( @{ $self->{finish} } ) {
+		$_ = $self->_make_token( @{ $_ } );
+	    }
+	} else {
+	    $self->{finish} = [
+		$self->_make_token( CLASS_DELIMITER, $end_delim ),
+	    ];
+	}
 
 	ref $_[1]
 	    and pos( $_[1] ) = pos $string;
@@ -263,6 +315,48 @@ sub delimiters {
 	qw{ start finish };
 }
 
+#	$self->_deprecation_notice( $type, $name );
+#
+#	This method centralizes deprecation. Type is 'attribute' or
+#	'method'. Deprecation is driven of the %deprecate hash. Values
+#	are:
+#	    false - no warning
+#	    1 - warn on first use
+#	    2 - warn on each use
+#	    3 - die on each use.
+#
+#	$self->_deprecation_in_progress( $type, $name )
+#
+#	This method returns true if the deprecation is in progress. In
+#	fact it returns the deprecation level.
+
+{
+
+    my %deprecate = (
+	attribute => {
+	    postderef	=> 3,
+	},
+    );
+
+    sub _deprecation_notice {
+	my ( undef, $type, $name, $repl ) = @_;		# Invocant unused
+	$deprecate{$type} or return;
+	$deprecate{$type}{$name} or return;
+	my $msg = sprintf 'The %s %s is %s', $name, $type,
+	    $deprecate{$type}{$name} > 2 ? 'removed' : 'deprecated';
+	defined $repl
+	    and $msg .= "; use $repl instead";
+	$deprecate{$type}{$name} >= 3
+	    and croak $msg;
+	warnings::enabled( 'deprecated' )
+	    and carp $msg;
+	$deprecate{$type}{$name} == 1
+	    and $deprecate{$type}{$name} = 0;
+	return;
+    }
+
+}
+
 sub _get_value_scalar {
     my ( $self, $method ) = @_;
     defined( my $val = $self->$method() )
@@ -319,11 +413,74 @@ sub handles {
     return $self->_stringify_source( $string, test => 1 );
 }
 
+sub indentation {
+    my ( $self ) = @_;
+    return $self->{indentation};
+}
+
 sub interpolates {
     my ( $self ) = @_;
     return $self->{interpolates};
 }
 
+sub location {
+    my ( $self ) = @_;
+    return $self->type()->location();
+}
+
+sub _make_token {
+    my ( $self, $class, $content, %arg ) = @_;
+    my $token = $class->__new( content => $content, %arg );
+    CLASS_UNKNOWN eq $class
+	and $self->{failures}++;
+    $self->{index_locations}
+	and $self->_update_location( $token );
+    return $token;
+}
+
+sub _update_location {
+    my ( $self, $token ) = @_;
+    $token->{location}	# Idempotent
+	and return;
+    my $loc = $self->{_location} ||= do {
+	my %loc = (
+	    line_content	=> '',
+	    location		=> $self->{location},
+	);
+	if ( __instance( $self->{source}, 'PPI::Element' ) ) {
+	    $loc{location} ||= $self->{source}->location();
+	    if ( my $doc = $self->{source}->document() ) {
+		$loc{tab_width} = $doc->tab_width();
+	    }
+	}
+	$loc{tab_width} ||= 1;
+	\%loc;
+    };
+    $loc->{location}
+	or return;
+    $token->{location} = [ @{ $loc->{location} } ];
+    if ( defined( my $content = $token->content() ) ) {
+	if ( my $newlines = $content =~ tr/\n/\n/ ) {
+	    $loc->{location}[LOCATION_LINE] += $newlines;
+	    $loc->{location}[LOCATION_LOGICAL_LINE] += $newlines;
+	    $content =~ s/ .* \n //smx;
+	    $loc->{location}[LOCATION_CHARACTER] =
+		$loc->{location}[LOCATION_COLUMN] = 1;
+	    $loc->{line_content} = '';
+	}
+	$loc->{location}[LOCATION_CHARACTER] += length $content;
+	$loc->{line_content} .= $content;
+	local $Text::Tabs::tabstop = $loc->{tab_width};
+	$loc->{location}[LOCATION_COLUMN] = 1 + length Text::Tabs::expand(
+	    $loc->{line_content} );
+    }
+    return;
+}
+
+sub parent {
+    return;
+}
+
 sub perl_version_introduced {
     my ( $self ) = @_;
     return List::Util::max( grep { defined $_ } MINIMUM_PERL,
@@ -348,6 +505,9 @@ sub perl_version_removed {
 
 sub postderef {
     my ( $self ) = @_;
+    # TODO postderef - eventually this goes away.
+    __PACKAGE__ eq caller
+	or $self->_deprecation_notice( attribute => 'postderef' );
     return $self->{postderef};
 }
 
@@ -379,6 +539,11 @@ sub start {
     return $self->{start}[ $inx || 0 ];
 }
 
+sub top {
+    my ( $self ) = @_;
+    return $self;
+}
+
 sub type {
     my ( $self, $inx ) = @_;
     $self->{type}
@@ -396,8 +561,6 @@ sub variables {
 
     my %var;
     foreach my $kid ( $self->children() ) {
-	$kid->can( 'variables' )
-	    or next;
 	foreach my $sym ( $kid->variables() ) {
 	    $var{$sym} = 1;
 	}
@@ -494,7 +657,6 @@ sub __decode {
 
 }
 
-
 # This subroutine was created in an attempt to simplify control flow.
 # Argument 2 (from 0) is not unpacked because the caller needs to see
 # the side effects of matches made on it.
@@ -503,77 +665,68 @@ sub __decode {
 
     my %special = (
 	'$$'	=> sub {	# Process ID.
-	    my ( $content ) = @_;
-	    return PPIx::QuoteLike::Token::Interpolation->__new(
-		content	=> $content,
-	    );
+	    my ( undef, $sigil ) = @_;
+	    return [ CLASS_INTERPOLATION, $sigil ];
 	},
 	'$'	=> sub {	# Called if we find (e.g.) '$@'
-	    my ( $content ) = @_;
-	    $_[1] =~ m/ \G ( [\@] ) /smxgc
+	    my ( undef, $sigil ) = @_;
+	    $_[2] =~ m/ \G ( [\@] ) /smxgc
 		or return;
-	    return PPIx::QuoteLike::Token::Interpolation->__new(
-		content	=> "$content$1",
-	    );
+	    return [ CLASS_INTERPOLATION, "$sigil$1" ];
 	},
 	'@'	=> sub {	# Called if we find '@@'.
-	    my ( $content ) = @_;
-	    return PPIx::QuoteLike::Token::String->__new(
-		content	=> $content,
-	    );
+	    my ( undef, $sigil ) = @_;
+	    return [ CLASS_STRING, $sigil ];
 	},
     );
 
     sub _interpolation {	## no critic (RequireArgUnpacking)
 	my ( $self, $sigil ) = @_;
+	# Argument $_[2] is $content, but we can't unpack it because we
+	# need the caller to see any changes to pos().
 
 	if ( $_[2] =~ m/ \G (?= \{ ) /smxgc ) {
-	    my $delim_re = _match_enclosed( qw< { > );
+	    # variable name enclosed in {}
+	    my $delim_re = __match_enclosed( qw< { > );
 	    $_[2] =~ m/ \G ( $delim_re ) /smxgc
-		and return PPIx::QuoteLike::Token::Interpolation->__new(
-		    content	=> "$sigil$1",
-		);
+		and return [ CLASS_INTERPOLATION, "$sigil$1" ];
 	    $_[2] =~ m/ \G ( .* ) /smxgc
-		and return $self->_unknown( "$sigil$1", MISMATCHED_DELIM );
+		and return [ CLASS_UNKNOWN, "$sigil$1",
+		    error => MISMATCHED_DELIM ];
 	    confess 'Failed to match /./';
 	}
 
 	if ( $_[2] =~ m< \G ( @{[ VARIABLE_RE ]} ) >smxgco
 	) {
+	    # variable name not enclosed in {}
 	    my $interp = "$sigil$1";
-	    my $deref = $self->postderef() ?
-		qr{ -> \@ | (?: -> )? }smx :
-		qr{ (?: -> )? }smx;
-	    while ( $_[2] =~ m/ \G  ( $deref ) (?= ( [[{] ) ) /smxgc ) { # }]
+	    while ( $_[2] =~ m/ \G  ( (?: -> )? ) (?= ( [[{] ) ) /smxgc ) { # }]
 		my $lead_in = $1;
-		my $delim_re = _match_enclosed( $2 );
+		my $delim_re = __match_enclosed( $2 );
 		if ( $_[2] =~ m/ \G ( $delim_re ) /smxgc ) {
 		    $interp .= "$lead_in$1";
 		} else {
 		    $_[2] =~ m/ ( .* ) /smxgc;
 		    return (
-			PPIx::QuoteLike::Token::Interpolation->__new(
-			    content	=> $interp,
-			),
-			$self->_unknown( "$1", MISMATCHED_DELIM ),
+			[ CLASS_INTERPOLATION, $interp ],
+			[ CLASS_UNKNOWN, "$1", error => MISMATCHED_DELIM ],
 		    );
 		}
 	    }
 
-	    # Postfix dereferencing
-	    $self->postderef()
-		and $_[2] =~ m/ \G ( -> (?: \$ \# | [\$\@] ) [*] ) /smxgc
-		and $interp .= $1;
+	    if ( $self->postderef()
+		    and defined( my $deref = _match_postderef( $_[2] ) ) ) {
+		$interp .= $deref;
+	    }
 
-	    return PPIx::QuoteLike::Token::Interpolation->__new(
-		content	=> $interp,
-	    );
+	    return [ CLASS_INTERPOLATION, $interp ];
 	}
 
 	my $code;
 	$code = $special{$sigil}
-	    and my $elem = $code->( $sigil, $_[2] )
-	    or return $self->_unknown( $sigil, 'Sigil without interpolation' );
+	    and my $elem = $code->( $self, $sigil, $_[2] )
+	    or return [ CLASS_UNKNOWN, $sigil,
+		error => 'Sigil without interpolation' ];
 
 	return $elem;
     }
@@ -601,78 +754,117 @@ sub _link_elems {
 }
 
 {
-    our %REGEXP_CACHE;
+    my %allow_subscr	= map { $_ => 1 } qw{ % @ };
 
-    my %matching_bracket = qw/ ( ) [ ] { } < > /;
+    # Match a postfix deref at the current position in the argument. If
+    # a match occurs it is returned, and the current position is
+    # updated. If not, nothing is returned, and the current position in
+    # the argument remains unchanged.
+    # This would all be much easier if I could count on Perl 5.10
+    sub _match_postderef {	## no critic (RequireArgUnpacking)
+	my $pos = pos $_[0];
+	$_[0] =~ m/ \G ( -> ) ( \$ \# | [\$\@%&*] ) /smxgc
+	    or return;
+	my $match = "$1$2";
+	my $sigil = $2;
+	$_[0] =~ m/ \G ( [*] ) /smxgc
+	    and return "$match$1";
 
-    sub _match_enclosed {
-	my ( $left ) = @_;
-	my $ql = quotemeta $left;
-	$REGEXP_CACHE{$ql}
-	    and return $REGEXP_CACHE{$ql};
-	if ( my $right = $matching_bracket{$left} ) {
+	if (
+	    $allow_subscr{$sigil} &&
+	    $_[0] =~ m/ \G (?= ( [[{] ) ) /smxgc	# }]
+	) {
+	    my $re = __match_enclosed( "$1" );
+	    $_[0] =~ m/ \G $re /smxgc
+		and return "$match$1";
+	}
 
-=begin comment
+	pos $_[0] = $pos;
+	return;
+    }
+}
 
-	    return ( $REGEXP_CACHE{$left} =
-		qr/ (
-		    \Q$left\E
-		    (?:
-			(?> [^\\\Q$left$right\E]+ ) |
-			(?> \$ [\Q$left$right\E] ) |
-			(?> \\ . ) |
-			(?-1)
-		    )*
-		    \Q$right\E
-		) /smx
-	    );
+# For various reasons we may get consecutive literals -- typically
+# strings. We want to merge these. The arguments are array refs, with
+# the class name of the token in [0] and the content in [1]. I know of
+# no way we can generate consecutive white space tokens, but if I did I
+# would want them merged.
+#
+# NOTE that merger loses all attributes of the second token, so we MUST
+# NOT merge CLASS_UNKNOWN tokens, or any class that might have
+# attributes other than content.
+{
+    my %can_merge = map { $_ => 1 } CLASS_STRING, CLASS_WHITESPACE;
 
-=end comment
-
-=cut
-
-	    # Based on Regexp::Common $RE{balanced} 2.113 (because I
-	    # can't use (?-1)
-
-	    my $ql = quotemeta $left;
-	    my $qr = quotemeta $right;
-	    my $pkg = __PACKAGE__;
-	    my $r  = "(??{ \$${pkg}::REGEXP_CACHE{'$ql'} })";
-
-	    my @parts = (
-		"(?>[^\\\\$ql$qr]+)",
-		"(?>\\\$[$ql$qr])",
-		'(?>\\\\.)',
-		$r,
-	    );
-
-	    {
-		use re qw{ eval };
-		local $" = '|';
-		$REGEXP_CACHE{$ql} = qr/($ql(?:@parts)*$qr)/;
+    sub _merge_strings {
+	my @arg = @_;
+	my @rslt;
+	foreach my $elem ( @arg ) {
+	    if ( @rslt && $can_merge{$elem->[0]}
+		&& $elem->[0] eq $rslt[-1][0]
+	    ) {
+		$rslt[-1][1] .= $elem->[1];
+	    } else {
+		push @rslt, $elem;
 	    }
+	}
+	return @rslt;
+    }
+}
 
-	    return $REGEXP_CACHE{$ql};
+# If we're processing an indented here document, strings must be split
+# on new lines and un-indented. We return array refs rather than
+# objects because we may be called before we're ready to build the
+# objects.
+sub _remove_here_doc_indentation {
+    my ( $self, $string, %arg ) = @_;
 
+    # NOTE that we rely on the fact that both undef (not indented) and
+    # '' (indented by zero characters) evaluate false.
+    $self->{indentation}
+	or return [ CLASS_STRING, $string ];
+
+    my $ignore_first;
+    if ( $arg{sibling} ) {
+	# Because the calling code primes the pump, @sibling will never
+	# be empty, even when processing the first token. So:
+	# * The pump-priming specifies class '', so if that is what we
+	#   see we must process the first line; otherwise
+	# * If the previous token is a string ending in "\n", we must
+	#   process the first line.
+	$ignore_first = '' ne $arg{sibling}[-1][0] && (
+	    CLASS_STRING ne $arg{sibling}[-1][0] ||
+	    $arg{sibling}[-1][1] !~ m/ \n \z /smx );
+    } else {
+	# Without @sibling, we unconditionally process the first line.
+	$ignore_first = 0;
+    }
+
+    my @rslt;
+
+    foreach ( split qr/ (?<= \n ) /smx, $string ) {
+	if ( $ignore_first ) {
+	    push @rslt, [ CLASS_STRING, "$_" ];
+	    $ignore_first = 0;
 	} else {
-
-	    # Based on Regexp::Common $RE{delimited}{-delim=>'`'}
-	    return ( $REGEXP_CACHE{$ql} ||=
-		qr< (?:
-		    (?: \Q$left\E )
-		    (?: [^\\\Q$left\E]* (?: \\ . [^\\\Q$left\E]* )* )
-		    (?: \Q$left\E )
-		) >smx
-	    );
+	    if ( "\n" eq $_ ) {
+		push @rslt,
+		    [ CLASS_STRING, "$_" ],
+		    ;
+	    } elsif ( s/ ( $self->{_indentation_re} ) //smx ) {
+		push @rslt,
+		    [ CLASS_WHITESPACE, "$1" ],
+		    [ CLASS_STRING, "$_" ],
+		    ;
+	    } else {
+		push @rslt,
+		    [ CLASS_UNKNOWN, "$_", error => NO_INDENTATION ],
+		    ;
+	    }
 	}
     }
 
-    sub _matching_delimiter {
-	my ( $left ) = @_;
-	my $right = $matching_bracket{$left}
-	    or return $left;
-	return $right;
-    }
+    return @rslt;
 }
 
 sub _stringify_source {
@@ -726,15 +918,6 @@ sub _stringify_source {
     return;
 }
 
-sub _unknown {
-    my ( $self, $content, $error ) = @_;
-    $self->{failures}++;
-    return PPIx::QuoteLike::Token::Unknown->__new(
-	content	=> $content,
-	error	=> $error,
-    );
-}
-
 sub _unquote {
     my ( $string ) = @_;
     $string =~ s/ \A ['"] //smx
@@ -767,6 +950,58 @@ reasonably like string literals. Its real reason for being is to find
 interpolated variables for L<Perl::Critic|Perl::Critic> policies and
 similar code.
 
+The parse is fairly straightforward, and a little poking around with
+F<eg/pqldump> should show how it normally goes.
+
+But there is at least one quote-like thing that probably needs some
+explanation.
+
+=head2 Indented Here Documents
+
+These were introduced in Perl 5.25.7 (November 2016) but not recognized
+by this module until its version 0.015 (February 2021). The indentation
+is parsed as
+L<PPIx::QuoteLike::Token::Whitespace|PPIx::Regexp::Token::Whitespace>
+objects, provided it is at least one character wide, otherwise it is not
+represented in the parse. That is to say,
+
+ <<~EOD
+     How doth the little crocodile
+     Improve his shining tail
+     EOD
+
+will have the three indentations represented by whitespace objects and
+each line of the literal represented by its own string object, but
+
+ <<~EOD
+ How doth the little crocodile
+ Improve his shining tail
+ EOD
+
+will parse the same as the non-indented version, except for the addition
+of the token representing the C<'~'>.
+
+L<PPI|PPI> is ahead of this module, and recognized indented here
+documents as of its version 1.246 (May 2019). Unfortunately, as of
+version 1.270 the indent gets lost in the parse, so a C<PPIx::QuoteLike>
+object initialized from such a
+L<PPI::Token::HereDoc|PPI::Token::HereDoc> will be seen as having an
+indentation of C<''> regardless of the actual indentation in the source.
+I believe this restriction will go away when
+L<https://github.com/Perl-Critic/PPI/issues/251> is resolved.
+
+=head1 DEPRECATION NOTICE
+
+The L<postderef|/postderef> argument to L<new()|/new> is being put
+through a deprecation cycle and retracted. After the retraction, postfix
+dereferences will always be recognized. This is the default behaviour
+now.
+
+Starting with version 0.012_01, the first use of this argument warned.
+With version 0.016_01, all uses will warn. With version 0.017_01 all
+uses will be fatal. With the first release after
+April 15 2022, all mention of this argument will be removed.
+
 =head1 INHERITANCE
 
 C<PPIx::QuoteLike> is not descended from any other class.
@@ -790,7 +1025,8 @@ quote-like literal of some sort, provided it begins like one. Otherwise
 this method will return nothing. The scalar representation of a here
 document is a multi-line string whose first line consists of the leading
 C< << > and the start delimiter, and whose subsequent lines consist of
-the content of the here document and the end delimiter.
+the content of the here document and the end delimiter. Indented here
+documents were not supported by this class until version C<0.015>.
 
 C<PPI> classes that can be handled are
 L<PPI::Token::Quote|PPI::Token::Quote>,
@@ -815,15 +1051,30 @@ If the C<$source> is a C<PPI::Element>, this encoding is used only if
 the document that contains the element has neither a byte order mark nor
 C<'use utf8'>.
 
+=item index_locations
+
+This Boolean argument determines whether the locations of the tokens
+should be computed. It defaults to true if the C<$source> argument is a
+L<PPI::Element|PPI::Element> or if the C<location> argument was
+provided, and false otherwise.
+
+=item location
+
+This argument is a reference to an array compatible with that returned
+by the L<PPI::Element|PPI::Element> location() method. It defaults to
+the location of the C<$source> argument if that was a
+L<PPI::Element|PPI::Element>, otherwise no locations will be available.
+
 =item postderef
 
+B<THIS ARGUMENT IS DEPRECATED>.
+See L<DEPRECATION NOTICE|/DEPRECATION NOTICE> above for the details.
+
 This Boolean argument determines whether postfix dereferencing is
 recognized in interpolation. If unspecified, or specified as C<undef>,
-it defaults to the value of C<$PPIx::QuoteLike::DEFAULT_POSTDEREF>. This
-variable is not exported, and is true by default. If you change the
-value, the change should be properly localized:
-
- local $PPIx::QuoteLike::DEFAULT_POSTDEREF = 0;
+it defaults to true. In version 0.012 it defaulted to the value of
+C<$PPIx::QuoteLike::DEFAULT_POSTDEREF>. This variable was not exported,
+and was true by default.
 
 =item trace
 
@@ -854,6 +1105,11 @@ This method returns all child elements. Children do not include the
 L<type()|/type>, or the L<start()|/start> or L<finish()|/finish>
 delimiters.
 
+=head2 column_number
+
+This method returns the column number of the first character in the
+element, or C<undef> if that can not be determined.
+
 =head2 content
 
  say $str->content();
@@ -931,6 +1187,15 @@ This convenience static method returns a true value if this package can
 be expected to handle the content of C<$string> (be it scalar or
 object), and a false value otherwise.
 
+=head2 indentation
+
+This method returns the indentation string if the object represents an
+indented here document, or C<undef> if it represents anything else,
+including an unindented here document.
+
+B<Note> that if indented syntax is used but the here document is not in
+fact indented, this will return C<''>, which evaluates to false.
+
 =head2 interpolates
 
  say $str->interpolates() ?
@@ -942,6 +1207,36 @@ a false value if it does not. This does B<not> indicate whether any
 interpolation actually takes place, only whether the string is
 double-quotish or single-quotish.
 
+=head2 line_number
+
+This method returns the line number of the first character in the
+element, or C<undef> if that can not be determined.
+
+=head2 location
+
+This method returns a reference to an array describing the position of
+the string, or C<undef> if the location is unavailable.
+
+The array is compatible with the corresponding
+L<PPI::Element|PPI::Element> method.
+
+=head2 logical_filename
+
+This method returns the logical file name (taking C<#line> directives
+into account) of the file containing first character in the element, or
+C<undef> if that can not be determined.
+
+=head2 logical_line_number
+
+This method returns the logical line number (taking C<#line> directives
+into account) of the first character in the element, or C<undef> if that
+can not be determined.
+
+=head2 parent
+
+This method returns nothing, since the invocant is only used at the top
+of the object hierarchy.
+
 =head2 perl_version_introduced
 
 This method returns the maximum value of C<perl_version_introduced>
@@ -994,6 +1289,30 @@ If called in list context you get the whole array. If called in scalar
 context you get the element whose index is given in the argument, or
 element zero if no argument is specified.
 
+=head2 statement
+
+This method returns the L<PPI::Statement|PPI::Statement> that
+contains this string, or nothing if the statement can not be
+determined.
+
+In general this method will return something only under the following
+conditions:
+
+=over
+
+=item * The string is contained in a L<PPIx::QuoteLike|PPIx::QuoteLike> object;
+
+=item * That object was initialized from a L<PPI::Element|PPI::Element>;
+
+=item * The L<PPI::Element|PPI::Element> is contained in a statement.
+
+=back
+
+=head2 top
+
+This method returns the top of the hierarchy -- in this case, the
+invocant.
+
 =head2 type
 
  my $type = $str->type();
@@ -1015,11 +1334,24 @@ or element zero if no argument is specified.
 
  say "Interpolates $_" for $str->variables();
 
+B<NOTE> that this method is discouraged, and may well be deprecated and
+removed. My problem with it is that it returns variable names rather
+than L<PPI::Element|PPI::Element> objects, leaving you no idea how the
+variables are used. It was originally written for the benefit of
+L<Perl::Critic::Policy::Variables::ProhibitUnusedVarsStricter|Perl::Critic::Policy::Variables::ProhibitUnusedVarsStricter>,
+but has proven inadequate to that policy's needs.
+
 This convenience method returns all interpolated variables. Each is
 returned only once, and they are returned in no particular order. If the
 object does not represent a string that interpolates, nothing is
 returned.
 
+=head2 visual_column_number
+
+This method returns the visual column number (taking tabs into account)
+of the first character in the element, or C<undef> if that can not be
+determined.
+
 =head1 RESTRICTIONS
 
 By the nature of this module, it is never going to get everything right.
@@ -1050,6 +1382,12 @@ things. I do not have any examples of where the parse of a quote-like
 thing would change based on what is interpolated, but neither can I rule
 it out. I<Caveat user>.
 
+=head2 PPI Restrictions
+
+As of version 0.015 of this module, the only known instance of this is
+the handling of indented here documents, as discussed above under
+L<Indented Here Documents|/Indented Here Documents>.
+
 =head2 Non-Standard Syntax
 
 There are modules out there that alter the syntax of Perl. If the syntax
@@ -1071,7 +1409,9 @@ L<Perl6::Pugs|Perl6::Pugs>. Enough said.
 =head1 SUPPORT
 
 Support is by the author. Please file bug reports at
-L<http://rt.cpan.org>, or in electronic mail to the author.
+L<https://rt.cpan.org/Public/Dist/Display.html?Name=PPIx-QuoteLike>,
+L<https://github.com/trwyant/perl-PPIx-QuoteLike/issues>, or in
+electronic mail to the author.
 
 =head1 AUTHOR
 
@@ -1079,7 +1419,7 @@ Thomas R. Wyant, III F<wyant at cpan dot org>
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright (C) 2016-2019 by Thomas R. Wyant, III
+Copyright (C) 2016-2021 by Thomas R. Wyant, III
 
 This program is free software; you can redistribute it and/or modify it
 under the same terms as Perl 5.10.0. For more details, see the full text
